/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: MainWindow.h
 *
 * Description: Main window header file for the graphical vi program.
 *
 **************************************************************************/


#ident "$Revision: 1.2 $"


#ifndef _MAINWINDOW_H
#define _MAINWINDOW_H


#include <Vk/VkApp.h>
#include <Vk/VkWindow.h>
#include <Vk/VkMenu.h>
#include "RecentFiles.h"
#include "ViWsh.h"
#include "Background.h"
#include "Catalog.h"


// Version information

#define GVI_VERSION	"gvi version 1.0.1"


/**************************************************************************
 *
 * Class: MainWindow
 *
 * Description: The application main window class for the gvi program.
 *
 **************************************************************************/

class MainWindow: public VkWindow
{
    friend class ViBackground;

    public:
	MainWindow(const char *name, ArgList args = NULL,
			Cardinal argCount = 0);
	~MainWindow();
	virtual const char *className();
	virtual void handleRawEvent(XEvent *event);
	virtual void afterRealizeHook();

    private:
	static VkMenuDesc fileMenu[];
	static VkMenuDesc selectedMenu[];
	static VkMenuDesc mainMenu[];

	VkMenuBar *menuBar;
	Catalog *catalog;
	Widget fileSelectionDialog;
	Widget newFileDialog;
	Widget lastFileDialog;
	Widget changeTabDialog;
	Widget recentFilesDialog;
	Dimension curWidth, curHeight;
	int deletePipe[2];
	ViWsh *deleteOnMap;
	RecentFiles *recentFiles;

	ViWsh* createViWsh(char *filename, Boolean update, Dimension width = 0,
				Dimension height = 0);
	void openFile(char *filename, Boolean update);
	void viMappedCB(VkCallbackObject *obj, void*, void*);
	void deleteOnMapCB(VkCallbackObject *obj, void*, void*);
	void viDestroyedCB(VkCallbackObject *obj, void*, void*);
	Boolean checkFileExists(const char *filename);
	static void deleteViCB(XtPointer client, int *fdp, XtInputId*);
	static void resizeCB(Widget, XtPointer client, XEvent *event);
	static void newCB(Widget, XtPointer client, XtPointer);
	static void openCB(Widget, XtPointer client, XtPointer);
	static void openRecentCB(Widget, XtPointer client, XtPointer);
	static void normalCloseCB(Widget, XtPointer client, XtPointer);
	static void saveCloseCB(Widget, XtPointer client, XtPointer);
	static void forceCloseCB(Widget, XtPointer client, XtPointer);
	static void lastFileClosedCB(Widget, XtPointer client, XtPointer);
	static void changeTabLabelCB(Widget, XtPointer client, XtPointer);
	static void tabLabelSelectedCB(Widget, XtPointer client,
						XtPointer callData);
	static void fileSelectedCB(Widget, XtPointer client,
						XtPointer callData);
	static void recentFilesSelectedCB(Widget, XtPointer client,
						XtPointer callData);
	static void newFileSelectedCB(Widget, XtPointer client,
						XtPointer callData);
	static void filenameVerifyCB(Widget, XtPointer, XtPointer callData);
	static void destroyDialogCB(Widget, XtPointer client, XtPointer);
	static void quitCB(Widget, XtPointer client, XtPointer);
	void quit();
	void resizeHandler();
};


/**************************************************************************
 *
 * Class: ViBackground
 *
 * Description: Brings up vi/wsh components as a background process.
 *
 **************************************************************************/

class ViBackground: public Background
{
    public:
	ViBackground() { fileIndex = 2; }

    protected:
	virtual Boolean timeSlice();
    
    private:
	int fileIndex;
};

#endif /* _MAINWINDOW_H */
