/**************************************************************************
 *
 * 		  Copyright (c)	1993 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: ProgramMsg.h
 *
 * Description: Header file for the program message printing class.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#ifndef _PROGRAMMSG_H
#define _PROGRAMMSG_H


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include "common.h"


/**************************************************************************
 *
 * Class: ProgramMsg
 *
 * Description: Provides a printf style interface to program error,
 *	warning and information message output. All messages are
 *	formatted as one of:
 *
 *	<program name> Error: <message>\n
 *	<program name> Warning: <message>\n
 *	<program name> Info: <message>\n
 *
 *	By default all messages are output to stderr.
 *
 **************************************************************************/

class ProgramMsg
{
    public:
    	ProgramMsg(const char *pname, FILE *fptr = stderr) {
	            ASSERT(pname != NULL);
	            ASSERT(fptr != NULL);
	            if ((progName = strdup(pname)) == NULL) {
		        fprintf(fptr, "No memory for program name\n");
		        exit(1);
		    }
	            outPtr = fptr;
	        }
	~ProgramMsg() { free(progName); }

	void error(const char *format, ...);
	void warn(const char *format, ...);
	void info(const char *format, ...);

	#ifdef DEBUG
	void debug(const char *format, ...);
	#endif

	void raw(const char *format, ...);
	void setOutputFile(FILE *fptr) {
		    ASSERT(fptr != NULL);
		    outPtr = fptr;
		}
	void setProgramName(const char *pname) {
		    ASSERT(pname != NULL);
		    free(progName);
		    if ((progName = strdup(pname)) == NULL) {
		        fprintf(outPtr, "No memory for program name\n");
		        exit(1);
		    }
		}
	char* getProgramName(void) { return progName; }

    private:
	void dispMsg(const char *format, va_list args);

	char *progName;			// Name of the current program
	FILE *outPtr;			// Output message file pointer
};


/* Declare a program scope global message class instance */

extern ProgramMsg progMsg;


#endif /* _PROGRAMMSG_H */
