/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: RecentFiles.h
 *
 * Description: Header for the RecentFiles class.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#ifndef _RECENTFILES_H
#define _RECENTFILES_H


#include <sys/types.h>


// Disk file info structure

typedef struct _fileInfo {
    char *pathname;			// Pathname of file
    char *label;			// Tab label
} FileInfo;


// File list structure

typedef struct _fileItem {
    FileInfo info;			// Disk file data
    _fileItem *next;			// Next file item
    _fileItem *prev;			// Last file item
} FileItem;


/**************************************************************************
 *
 * Class: RecentFiles
 *
 * Description: Manages a disk based list of the N most recently opened
 *	files. This allows a user to quickly open a set of the last files
 *	that were being edited.
 *
 **************************************************************************/

class RecentFiles
{
    public:
	RecentFiles();
	~RecentFiles();

	void update(const FileInfo *finfo);
	int getFilenames(char ***filenamesp);
	void freeFilenames(char **filenames);
	char* getLabel(const char *pathname);

    private:
	FileItem *fileList;		// Doubly linked list of recent files
	int maxFiles;			// Max number of remembered files
	char *diskPathname;		// Pathname of disk based list file
	time_t diskModTime;		// Time disk list last modified

	void dump();
	void add(const FileInfo *finfo);
	FileItem* remove(FileItem *item);
	void readList();
	void writeList();
};


#endif /* _RECENTFILES_H */
