/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: common.c++
 *
 * Description: Common utility routines.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "safeMemory.h"
#include "common.h"


/**************************************************************************
 *
 * Function: _rtassert
 *
 * Description: Runtime assertion function. Called by the ASSERT macro.
 *	This function does the same thing as assert but the macro does
 *	not get compiled out if NDEBUG is defined.
 *
 * Parameters: 
 *	exp (I) - expression that failed the assertion
 *	fname (I) - name of file in which the assertion failure occurred
 *	linenum (I) - line number in fname where assertion failed
 *
 * Return: none
 *
 **************************************************************************/

void _rtassert(const char *exp, const char *fname, int linenum)
{
    fprintf(stderr, "Assertion failed: %s, file %s, line %d\n",
			exp, fname, linenum);
    abort();
}


/**************************************************************************
 *
 * Function: makeAbsolutePathname
 *
 * Description: Turns the specified filename into an absolute pathname.
 *
 * Parameters:
 *      filename (I) - filename to convert. Can be NULL.
 *
 * Return: Pointer to absolute pathname or NULL if filename is NULL. If
 *      an absolute pathname could not be constructed the original filename
 *      is returned. Storage for the pathname should be freed with a call
 *      to free when no longer needed.
 *
 **************************************************************************/

char* makeAbsolutePathname(const char *filename)
{
    char *path;
    char *cwd;
    int needFree = 0;
    size_t len;

    if (filename == NULL)
        return NULL;

    if (*filename == '/')
        return safeStrdup(filename);

    if ((cwd = getenv("PWD")) == NULL) {
        if ((cwd = getcwd(NULL, PATH_MAX + 1)) == NULL)
            return safeStrdup(filename);
        needFree = 1;
    }
    len = strlen(cwd);
    path = (char*)safeMalloc(len + strlen(filename) + 5);
    if (cwd[len-1] == '/')
        sprintf(path, "%s%s", cwd, filename);
    else
        sprintf(path, "%s/%s", cwd, filename);
    if (needFree)
        free(cwd);

    return path;
}
