/*
 * Copyright 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
//
//  Class:		ConfDoc
//  Base Classes:       none
//  Author:             Jesse Rendleman (jesse@sgi.com)
//

// conf includes
#include "ConfMsg.h"
#include "ConfList.h"
#include "ConfText.h"

// unix includes
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <netinet/in.h>

// X11/motif includes
#include <Xm/Xm.h>

#define XCONF_MCAST_GROUP  "225.0.0.250"
#define MAX_MCAST_HOPS  31

#define XCONF_RC_FILE  "~/.xconfrc"

#define DEFPORT 6090


class ConfDoc {
  private:
  protected:
    XtAppContext _app;

    char *_group;
    char *_user;
    char* _server;

    // currently not used
    // GroupList _groupList;
    ConfList _groupList;
    ConfList _userList;
    ConfList _helpList;

    Widget _inputLabel;

    Widget _sendText;
    Widget _recvText;
    ConfText *_sendConfText;
    ConfText *_recvConfText;

    int _port;
    int _socket;
    struct sockaddr_in _sendAddr;
    struct sockaddr_in _recvAddr;
    int _addrSize;

    Boolean setupSockets();
    Boolean setupUDP();
    Boolean setupTCP();

    Boolean setName(char *);
    Boolean sendMsg(short, char *, char *, char *);
    Boolean sendBroadcast(char *);
    Boolean sendUrgent(char *);

    void setInputLabel(char *);

    char *getWord(char *, int);
    int numWords(char *buffer);

    static void helpCB(Widget, XtPointer, XtPointer);
    static void sendCB(Widget, XtPointer, XtPointer);
    static void usersCB(Widget, XtPointer, XtPointer);
    static void groupsCB(Widget, XtPointer, XtPointer);
    static void recvCB(XtPointer, int, XtInputId *);

  public:
    ConfDoc(XtAppContext app, const char *name);
    ~ConfDoc();
    void init(int port, char* server);
    Widget build(Widget parent);
    Boolean setGroup(char *);
    Boolean setupChannel(char *);
    Boolean findUser(char *);
    Boolean sendPrivate(char *, int);

    // should be public, so main() can call it to handle window mgr signals
    static void quitCB(Widget, XtPointer, XtPointer);
};

